#!/usr/bin/env python3
import time, struct, gzip, threading, queue
import readchar
from tinyrv import sim, zext, sext

# Heavily based on https://github.com/cnlohr/mini-rv32ima Big thanks to CNLohr!
class rvlinux(sim):
    def __init__(self, image, dtb, ram_size, command_line):
        super().__init__(xlen=32, trap_misaligned=False)
        self.ram_size = ram_size
        self.ram_base = 0x8000_0000

        # load kernel image
        self.copy_in(image, self.ram_base)

        # load and patch DTB
        dtb_addr = self.ram_base+self.ram_size
        self.copy_in(dtb, dtb_addr)
        struct.pack_into('>I', *self.page_and_offset(dtb_addr+0x13c), ram_size)
        struct.pack_into('54s', *self.page_and_offset(dtb_addr+0xc0), command_line.encode())

        # set up initial vm state
        self.pc = self.ram_base
        self.x[self.a0] = 0  # hart ID
        self.x[self.a1] = dtb_addr
        self.wfi = False
        self.flux_capacitor = 0  # for time travel

        # start thread for capturing user key presses
        self.input = queue.Queue()
        def add_input(input):
            while True: input.put(readchar.readchar())
        threading.Thread(target=add_input, args=(self.input,), daemon=True).start()

    def copy_in(self, bytes, base=0):
        for po in range(0, len(bytes), self.mem_psize):
            page, pa = self.page_and_offset(base+po)
            assert pa == 0, f"base must be aligned to page size ({self.mem_psize})."
            nbytes = min(len(bytes)-po, self.mem_psize)
            page[0:nbytes] = bytes[po:po+nbytes]

    def hook_csr(self, csr, reqval):
        if csr == 0x139: char = chr(reqval); print(char, end='')  # console output
        elif csr == 0x140: return super().hook_csr(csr, -1 if self.input.empty() else ord(self.input.get()))  # console input
        return super().hook_csr(csr, reqval)

    def timer_fired(self):
        # multiplier should actually be *1_000_000. clock runs 1000x slower than real-time to avoid kernel panic.
        current_time = zext(64, int(time.perf_counter()*1000 + self.flux_capacitor))
        match_time = self.load('Q', 0x1100_4000, notify=False)
        self.store('Q', 0x1100bff8, current_time, notify=False)
        if current_time >= match_time: self.csr[self.mip] |= 1<<7  # timer expired
        else: self.csr[self.mip] &= ~(1<<7)
        if (self.csr[self.mstatus]&0x8) and (self.csr[self.mie]&self.csr[self.mip]&(1<<7)):
            self.mtrap(0, sext(32, 0x8000_0007))
            self.current_mode = 3
            self.wfi = False
            return True
        return False

    def hook_exec(self):
        if (self.cycle % 1000) == 0 and self.timer_fired(): return False  # don't execute current op
        while self.wfi:
            time.sleep(0.005); self.flux_capacitor += 5000  # when waiting, no harm in running clock close to real-time.
            if self.timer_fired(): return False  # done waiting. don't execute current op
        return super().hook_exec()  # continue normally

    def _wfi(self, **_): self.pc += 4; self.csr[self.mstatus] |= 0x8; self.wfi = True

if __name__ == '__main__':
    linux = rvlinux(gzip.open('Image.gz', 'rb').read(), open('sixtyfourmb.dtb', 'rb').read(), ram_size=64*1024*1024, command_line='console=hvc0')
    print('booting linux...')
    linux.run(0, trace=False)


#!/usr/bin/env python3
import re, csv, yaml, pathlib, collections

try:
    def dr(h,l): return list(range(h,l-1,-1))
    base = pathlib.Path('riscv-opcodes')
    opcodes = yaml.safe_load(open(base / 'instr_dict.yaml'))
    csrs = dict((int(a, 16), n) for fn in ['csrs.csv', 'csrs32.csv'] for a, n in csv.reader(open(base / fn), skipinitialspace=True))
    arg_bits = dict((a, dr(int(h),int(l))) for a, h, l in csv.reader(open(base / 'arg_lut.csv'), skipinitialspace=True))
    for s in open(base / 'constants.py').readlines():  # immediate scrambling from latex_mapping. Some better way?
        if mask := re.match(r"latex_mapping\[['\"](.*?)['\"]\] = ['\"][^\[]*\[([^\]]*)\]['\"]", s):
            fbits = sum([(dr(*(int(i) for i in part.split(':'))) if ':' in part else [int(part)]) for part in mask[2].split('$\\\\vert$')], [])
            locs = [-1] * (max(fbits)+1)
            for i, b in enumerate(fbits): locs[-b-1] = arg_bits[mask[1]][i]
            arg_bits[mask[1]] = locs #[31] * (32-len(locs)) + locs if locs[0] == 31 else locs  # sign extension to 32 bits
except Exception as e: raise Exception("Unable to load RISC-V specs. Do:\n"
                                       "git clone https://github.com/riscv/riscv-opcodes.git\n"
                                       "make -C riscv-opcodes")

for name, op in opcodes.items():
    op['name'], op['mask'], op['match'] = name, int(op['mask'], 16), int(op['match'], 16)
    del op['encoding']
    op['arg_bits'] = {}
    for vf in op['variable_fields']:
        if vf not in arg_bits: continue
        bits = [-1] * (32-len(arg_bits[vf])) + arg_bits[vf]  # padding for proper alignment when combining hi and lo fields
        vf2 = vf.replace('hi','').replace('lo','').replace('c_','')
        op['arg_bits'][vf2] = [max(a,b) for a, b in zip(op['arg_bits'][vf2], bits)] if vf2 in op['arg_bits'] else bits
    for vf in op['arg_bits']: op['arg_bits'][vf] = op['arg_bits'][vf][next(n for n, e in enumerate(op['arg_bits'][vf]) if e >= 0):]  # remove padding

common_ops = ('addi,sw,lw,jal,bne,beq,add,jalr,lbu,slli,lui,andi,or,bltu,srli,and,sub,blt,bgeu,xor,sb,auipc,sltiu,bge,lb,mul,sltu,lhu,sll,srl,sh,amoadd_w,xori,ori,csrrci,csrrs,srai,fence,lr_w,sc_w,mulhu,amoor_w,lh,amoand_w,csrrsi,divu,div,remu,sra,slt,csrrw,amoswap_w,csrrc,mulh,fence_i,rem,mret,csrrwi').split(',')

mask_match = []
mask_match_aliases = collections.defaultdict(set)
for mask in dict((opcodes[op]['mask'],1) for op in common_ops + list(opcodes)):
    matches = {}
    for op in opcodes.values():
        if op['mask'] != mask or op['name'].endswith('_rv32'): continue
        if op['match'] in matches: mask_match_aliases[matches[op['match']][1:-1]].add(op['name'])
        else: matches[op['match']] = f'${op["name"]}$'
    mask_match.append((mask, matches))

print('writing tinyrv/opcodes.py')
with open('tinyrv/opcodes.py', 'w') as f:
    f.write(f'# auto-generated by tinyrv_opcodes_gen.py\n')
    f.write(f'opcodes={str(opcodes)}\n')
    f.write(f'arg_bits={str(arg_bits)}\n')
    f.write(f'csrs={str(csrs)}\n')
    dstr = str(mask_match).replace("'$", "opcodes['").replace("$'", "']")
    f.write(f'mask_match={dstr}\n')
    f.write(f'mask_match_aliases={dict(mask_match_aliases)}\n')


#!/usr/bin/env python3
import sys, pathlib, tinyrv

def main():
    if len(sys.argv) < 2: print(f'usage: {pathlib.Path(sys.argv[0]).name} file.bin [(32|64) [run limit in decimal [start PC in hex]]]'); exit(1)
    xlen = 32 if len(sys.argv) < 3 else int(sys.argv[2])
    limit = 0 if len(sys.argv) < 4 else int(sys.argv[3])
    pc = 0 if len(sys.argv) < 5 else int(sys.argv[4], 16)
    rv = tinyrv.sim(xlen=xlen, trap_misaligned=False)
    rv.read_bin(sys.argv[1], base=0)
    rv.pc = pc
    rv.run(limit)

if __name__ == '__main__': main()


#!/usr/bin/env python3
import sys, subprocess
from tinyrv import sim

xlen = int(sys.argv[1])

subprocess.run('riscv64-unknown-elf-objcopy -O binary my.elf my.bin'.split(' '))

syms = dict((fields[-1], int(fields[0],16)) for fields in [line.decode().strip().split(' ') for line in subprocess.Popen(['riscv64-unknown-elf-objdump', '-t', 'my.elf'], stdout=subprocess.PIPE).stdout.readlines()] if len(fields) > 2 and fields[1] in ('l', 'g'))

rv = sim(xlen=xlen)
rv.read_bin('my.bin', base=syms['.text.init'])
rv.pc = syms['.text.init']
rv.run(16000, bpts={syms['rvtest_code_end']})

with open('DUT-tinyrv.signature', 'w') as f:
    for addr in range(syms['begin_signature'], syms['end_signature'], 4):
        f.write(f'{rv.load("I",addr):08x}\n')
    if rv.op.addr != syms['rvtest_code_end']:
        print(f'test unfinished? {rv.op.addr:016x} {syms["rvtest_code_end"]:016x}')



#!/usr/bin/env python3
import subprocess
from pathlib import Path

from tinyrv import sim

# git clone https://github.com/riscv-software-src/riscv-tests.git
# cd riscv-tests
# autoconf
# ./configure
# make

class rvsim2(sim):
    def _ecall(self, **_):
        if self.x[self.a7] == 93 and self.x[self.a0] == 0:
            self.passed = True
        else:
            print(f'test failed ({self.x[self.a0]})')

for f in Path('riscv-tests/isa').glob('rv??ui-p-*'):
    if '.' in str(f): continue
    subprocess.run(f'riscv64-unknown-elf-objcopy -O binary {str(f)} {str(f)}.bin'.split(' ')) 
    print('running', f)
    rv = rvsim2(xlen=64 if 'rv64' in str(f) else 32, trap_misaligned=False)
    rv.read_bin(str(f)+'.bin', base=0)
    rv.passed = False
    rv.run(2000, trace=False)
    if not rv.passed:
        print(f'test failed!!! {f}')
        exit(1)



#!/usr/bin/env python3
import sys, pathlib, tinyrv

def main():
    if len(sys.argv) < 2: print(f'usage {pathlib.Path(sys.argv[0]).name} (file.bin | hex [hex ...])'); exit(1)
    for op in tinyrv.decoder(*sys.argv[1:]):
        print(f'{tinyrv.zext(64,op.addr):08x}: {str(op):40} # {", ".join(op.extension) if op.valid() else "INVALID data=" + hex(op.data)}')

if __name__ == '__main__': main()


import os
import logging

import riscof.utils as utils
from riscof.pluginTemplate import pluginTemplate

logger = logging.getLogger()

class tinyrv(pluginTemplate):
    __model__ = "tinyrv"
    __version__ = "latest"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = kwargs['config']
        self.pluginpath=os.path.abspath(config['pluginpath'])
        self.dut_exe = "python3 " + os.path.join(self.pluginpath, "runner.py")
        self.num_jobs = str(config.get('jobs', 1))
        self.isa_spec = os.path.abspath(config['ispec'])
        self.platform_spec = os.path.abspath(config['pspec'])
        self.target_run = config.get('target_run', '1') != '0'

    def initialise(self, suite, work_dir, archtest_env):
       self.work_dir = work_dir
       self.suite_dir = suite
       self.compile_cmd = 'riscv{1}-unknown-elf-gcc -march={0} \
         -static -mcmodel=medany -fvisibility=hidden -nostdlib -nostartfiles -g\
         -T '+self.pluginpath+'/env/link.ld\
         -I '+self.pluginpath+'/env/\
         -I ' + archtest_env + ' {2} -o {3} {4}'

    def build(self, isa_yaml, platform_yaml):
      ispec = utils.load_yaml(isa_yaml)['hart0']
      self.xlen = ('64' if 64 in ispec['supported_xlen'] else '32')
      self.compile_cmd = self.compile_cmd+' -mabi='+('lp64 ' if 64 in ispec['supported_xlen'] else 'ilp32 ')

    def runTests(self, testList):

      # Delete Makefile if it already exists.
      if os.path.exists(self.work_dir+ "/Makefile." + self.name[:-1]):
            os.remove(self.work_dir+ "/Makefile." + self.name[:-1])
      
      make = utils.makeUtil(makefilePath=os.path.join(self.work_dir, "Makefile." + self.name[:-1]))
      make.makeCommand = 'make -k -j' + self.num_jobs

      for testname in testList:
          testentry = testList[testname]
          test = testentry['test_path']
          test_dir = testentry['work_dir']
          elf = 'my.elf'
          compile_macros= ' -D' + " -D".join(testentry['macros'])
          cmd = self.compile_cmd.format(testentry['isa'].lower(), self.xlen, test, elf, compile_macros)
          if self.target_run:
            simcmd = self.dut_exe + (' 64' if '64' in testentry['isa'] else ' 32')
          else:
            simcmd = 'echo "NO RUN"'

          execute = '@cd {0}; {1}; {2};'.format(testentry['work_dir'], cmd, simcmd)
          make.add_target(execute)

      make.execute_all(self.work_dir)
      if not self.target_run:
          raise SystemExit(0)



import os, struct, array, struct
from .opcodes import *

iregs = 'zero,ra,sp,gp,tp,t0,t1,t2,fp,s1,a0,a1,a2,a3,a4,a5,a6,a7,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,t3,t4,t5,t6'.split(',')
customs={0b0001011: 'custom0', 0b0101011: 'custom1', 0b1011011: 'custom2', 0b1111011: 'custom3'}
def zext(length, word): return word&((1<<length)-1)
def sext(length, word): return word|~((1<<length)-1) if word&(1<<(length-1)) else zext(length, word)
def xfmt(length, word): return f'{{:0{length//4}x}}'.format(zext(length, word))

class rvop:
    def __init__(self, **kwargs): [setattr(self, k, v) for k, v in kwargs.items()]
    def arg_str(self):
        if self.name in 'lb,lh,lw,ld,lbu,lhu,lwu,sb,sh,sw,sd,jalr'.split(','): args = [iregs[self.rd] if 'rd' in self.args else iregs[self.rs2], f"{self.imm12}({iregs[self.rs1]})"]
        elif self.name[:3] == 'csr': args = [iregs[self.rd], csrs.get(self.csr, hex(self.csr)), iregs[self.rs1] if 'rs1' in self.args else f'{hex(self.zimm) if abs(self.zimm) > 255 else self.zimm}']
        elif self.name[:5] == 'fence': args = [''.join(c for c, b in zip([*'iorw'], [i=='1' for i in f'{self.args[name]:04b}']) if b) for name in ('pred', 'succ') if name in self.args] + [f'fm={self.fm:04b}' if 'fm' in self.args else None]
        elif (self.name[:2] == 'c_') or (self.name[0] == 'f'): args = [f'{k}={v}' for k, v in self.args.items()]  # TODO: compressed and fp ops
        elif self.name == 'jal' or ('bimm12' in self.args): args = [iregs[self.rd] if 'rd' in self.args else None, iregs[self.rs1] if 'rs1' in self.args else None, iregs[self.rs2] if 'rs2' in self.args else None, hex(zext(64, self.addr+(self.jimm20 if 'jimm20' in self.args else self.bimm12)))]
        elif ('rd' in self.args) and ('rs1' in self.args): args = [iregs[self.rd], iregs[self.rs1], iregs[self.rs2] if 'rs2' in self.args else None, *[f'{hex(self.args[name]) if abs(self.args[name]) > 255 else self.args[name]}' for name in ('imm12','shamtw','shamtd') if name in self.args]]
        elif ('rd' in self.args) and ('imm20' in self.args): args = [iregs[self.rd], hex(zext(32,self.imm20) if 'l' in self.name else self.imm20) if abs(self.imm20) > 255 else f'{self.imm20}']
        else: args = [f'{k}={v}' for k, v in self.args.items()]  # fallback
        return ', '.join([a for a in args if a is not None])
    def valid(self): return min([not('nz' in k or 'n0' in k) or v!=0 for k, v in self.args.items()] + [hasattr(self, 'extension')])
    def __repr__(self): return f'{self.name.replace("_","."):10} {self.arg_str()}'

def rvsplitter(*data, base=0, lower16=0):  # yields addresses and 32-bit/16-bit(compressed) RISC-V instruction words.
    for addr, instr in enumerate(struct.iter_unpack('<H', open(data[0],'rb').read() if isinstance(data[0],str) and os.path.isfile(data[0]) else array.array('I',[int(d,16) if isinstance(d,str) else d for d in (data[0] if hasattr(data[0], '__iter__') and not isinstance(data[0],str) else data)]))):
        if lower16: yield int(base)+(addr-1)*2, (instr[0]<<16)|lower16; lower16 = 0
        elif instr[0]&3 == 3: lower16 = instr[0]  # Two LSBs set: 32-bit instruction
        else: yield int(base)+addr*2, instr[0]

def decode(instr, addr=0):  # decodes one instruction
    o = rvop(addr=addr, data=instr, name=customs.get(instr&0b1111111,'UNKNOWN'), args={})
    for mask, m_dict in mask_match:
        if op := m_dict.get(instr&mask, None):
            for vf, bits in op['arg_bits'].items():
                value = 0
                for bit in bits: value = (value<<1) | ((instr>>bit)&1 if bit>=0 else 0)
                o.args[vf] = sext(len(bits), value) if bits[0] == 31 else value
            [setattr(o,k,v) for k,v in (op|o.args).items()]
            break
    return o

def decoder(*data, base=0):  # yields decoded instructions.
    for addr, instr in rvsplitter(*data, base=base):
        if instr != 0: yield decode(instr, addr)

class sim:  # simulates RV32IMAZicsr_Zifencei, RV64IMAZicsr_Zifencei
    class rvregs:
        def __init__(self, xlen, sim): self._x, self.xlen, self.sim = [0]*32, xlen, sim
        def __getitem__(self, i): return self._x[i]
        def __setitem__(self, i, d):
            if i!=0 and (self.sim.trace_log is not None) and d!=self._x[i]: self.sim.trace_log.append(f'{iregs[i]}=' + (f'{zext(self.xlen, d):08x}' if d>>32 in (0, -1) else f'{zext(self.xlen, d):016x}'))
            if i!=0: self._x[i] = d
    def __init__(self, xlen=64, trap_misaligned=True):
        self.xlen, self.trap_misaligned, self.trace_log = xlen, trap_misaligned, []
        self.pc, self.x, self.f, self.csr, self.lr_res_addr, self.cycle, self.current_mode, self.mem_psize, self.mem_pages, self.fmt_conv = 0, self.rvregs(self.xlen, self), [0]*32, [0]*4096, -1, 0, 3, 2<<12, {}, {'q': 64, 'Q': 64, 'i': 32, 'I': 32, 'h': 16, 'H': 16, 'b': 8, 'B': 8, 32: 'i', 64: 'q'}
        [setattr(self, n, i) for i, n in list(enumerate(iregs))+list(csrs.items())]  # convenience
    def __repr__(self): return '\n'.join(['  '.join([f'x{r+rr:02d}({(iregs[r+rr])[-2:]})={xfmt(self.xlen, self.x[r+rr])}' for r in range(0, 32, 8)]) for rr in range(8)])
    def hook_csr(self, csr, reqval): return reqval if (csr&0xc00)!=0xc00 else self.csr[csr]
    def notify_stored(self, addr): pass  # called *after* mem store
    def notify_loading(self, addr): pass  # called *before* mem load
    def mtrap(self, tval, cause):
        self.csr[self.mtval], self.csr[self.mepc], self.csr[self.mcause], self.pc = zext(self.xlen,tval), self.op.addr, cause, zext(self.xlen,self.csr[self.mtvec]&(~3))
        if self.trace_log is not None: self.trace_log.append(f'mtrap cause={hex(cause)} tval={hex(tval)}')
        self.csr[self.mstatus] = ((self.csr[self.mstatus]&0x08) << 4) | (self.current_mode << 11)
    def read_bin(self, file, base=0): [self.store('B', i+base, b, notify=False) for i, b in enumerate(open(file, 'rb').read())]  # FIXME: SLOOOOW
    def page_and_offset(self, addr):
        if addr&~(self.mem_psize-1) not in self.mem_pages: self.mem_pages[addr&~(self.mem_psize-1)] = bytearray(self.mem_psize)
        return self.mem_pages[addr&~(self.mem_psize-1)], addr&(self.mem_psize-1)
    def store(self, format, addr, data, notify=True):
        if self.trace_log is not None: self.trace_log.append(f'{xfmt(struct.calcsize(format)*8, data)}->mem[{xfmt(self.xlen, addr)}]')
        if self.trap_misaligned and addr&(struct.calcsize(format)-1) != 0: self.mtrap(addr, 6)
        else: struct.pack_into(format, *self.page_and_offset(zext(self.xlen,addr)), data)
        if notify: self.notify_stored(zext(self.xlen,addr))
    def load(self, format, addr, fallback=0, notify=True, check_misaligned=True):
        if check_misaligned and self.trap_misaligned and addr&(struct.calcsize(format)-1) != 0: self.mtrap(addr, 4); return fallback
        if notify: self.notify_loading(zext(self.xlen,addr))
        data = struct.unpack_from(format, *self.page_and_offset(zext(self.xlen,addr)))[0]
        if self.trace_log is not None: self.trace_log.append(f'mem[{xfmt(self.xlen, addr)}]->{xfmt(struct.calcsize(format)*8, data)}')
        return data
    def idiv2zero(self, a, b): return -(-a // b) if (a < 0) ^ (b < 0) else a // b
    def rem2zero(self, a, b): return a - b * self.idiv2zero(a, b)
    def _auipc    (self, rd, imm20,        **_): self.pc+=4; self.x[rd] = sext(self.xlen, self.op.addr+imm20)
    def _lui      (self, rd, imm20,        **_): self.pc+=4; self.x[rd] = imm20
    def _jal      (self, rd, jimm20,       **_): self.pc, self.x[rd] = zext(self.xlen, self.op.addr+jimm20),    self.op.addr+4
    def _jalr     (self, rd, rs1, imm12,   **_): self.pc, self.x[rd] = zext(self.xlen, self.x[rs1]+imm12)&(~1), self.op.addr+4 # LSB=0
    def _beq      (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if self.x[rs1] == self.x[rs2] else self.op.addr+4
    def _bne      (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if self.x[rs1] != self.x[rs2] else self.op.addr+4
    def _blt      (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if self.x[rs1] <  self.x[rs2] else self.op.addr+4
    def _bge      (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if self.x[rs1] >= self.x[rs2] else self.op.addr+4
    def _bltu     (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if zext(self.xlen, self.x[rs1]) <  zext(self.xlen, self.x[rs2]) else self.op.addr+4
    def _bgeu     (self, rs1, rs2, bimm12, **_): self.pc = (self.op.addr+bimm12) if zext(self.xlen, self.x[rs1]) >= zext(self.xlen, self.x[rs2]) else self.op.addr+4
    def _sb       (self, rs1, rs2, imm12,  **_): self.pc+=4; self.store('B', self.x[rs1]+imm12, zext( 8,self.x[rs2]))
    def _sh       (self, rs1, rs2, imm12,  **_): self.pc+=4; self.store('H', self.x[rs1]+imm12, zext(16,self.x[rs2]))
    def _sw       (self, rs1, rs2, imm12,  **_): self.pc+=4; self.store('I', self.x[rs1]+imm12, zext(32,self.x[rs2]))
    def _sd       (self, rs1, rs2, imm12,  **_): self.pc+=4; self.store('Q', self.x[rs1]+imm12, zext(64,self.x[rs2]))
    def _lb       (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('b', self.x[rs1]+imm12, self.x[rd])
    def _lbu      (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('B', self.x[rs1]+imm12, self.x[rd])
    def _lh       (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('h', self.x[rs1]+imm12, self.x[rd])
    def _lw       (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('i', self.x[rs1]+imm12, self.x[rd])
    def _ld       (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('q', self.x[rs1]+imm12, self.x[rd])
    def _lhu      (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('H', self.x[rs1]+imm12, self.x[rd])
    def _lwu      (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = self.load('I', self.x[rs1]+imm12, self.x[rd])
    def _addi     (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  +          imm12)
    def _xori     (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  ^          imm12)
    def _ori      (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  |          imm12)
    def _andi     (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  &          imm12)
    def _addiw    (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = sext(32,          self.x[rs1]  +          imm12)
    def _addw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, sext(32, self.x[rs1]) + sext(32, self.x[rs2]))
    def _subw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, sext(32, self.x[rs1]) - sext(32, self.x[rs2]))
    def _add      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  +          self.x[rs2])
    def _sub      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  -          self.x[rs2])
    def _xor      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  ^          self.x[rs2])
    def _or       (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  |          self.x[rs2])
    def _and      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,   self.x[rs1]  &          self.x[rs2])
    def _sltiu    (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] = zext(self.xlen,   self.x[rs1]) < zext(self.xlen, imm12)
    def _sltu     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = zext(self.xlen,   self.x[rs1]) < zext(self.xlen, self.x[rs2])
    def _slti     (self, rd, rs1, imm12,   **_): self.pc+=4; self.x[rd] =                   self.x[rs1]  <                 imm12
    def _slt      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] =                   self.x[rs1]  <                 self.x[rs2]
    def _slliw    (self, rd, rs1, shamtw,  **_): self.pc+=4; self.x[rd] = sext(32,                        self.x[rs1]  << shamtw)
    def _srliw    (self, rd, rs1, shamtw,  **_): self.pc+=4; self.x[rd] = sext(32,        zext(32,        self.x[rs1]) >> shamtw)
    def _sraiw    (self, rd, rs1, shamtw,  **_): self.pc+=4; self.x[rd] = sext(32,        sext(32,        self.x[rs1]) >> shamtw)
    def _slli     (self, rd, rs1, shamtd,  **_): self.pc+=4; self.x[rd] = sext(self.xlen,                 self.x[rs1]  << shamtd)  # shared with RV64I
    def _srai     (self, rd, rs1, shamtd,  **_): self.pc+=4; self.x[rd] = sext(self.xlen,                 self.x[rs1]  >> shamtd)  # shared with RV64I
    def _srli     (self, rd, rs1, shamtd,  **_): self.pc+=4; self.x[rd] = sext(self.xlen, zext(self.xlen, self.x[rs1]) >> shamtd)  # shared with RV64I
    def _sll      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,                 self.x[rs1]  << (self.x[rs2]&(self.xlen-1)))
    def _srl      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, zext(self.xlen, self.x[rs1]) >> (self.x[rs2]&(self.xlen-1)))
    def _sra      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen,                 self.x[rs1]  >> (self.x[rs2]&(self.xlen-1)))
    def _sllw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32,        sext(32,        self.x[rs1]) << (self.x[rs2]&31))
    def _srlw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32,        zext(32,        self.x[rs1]) >> (self.x[rs2]&31))
    def _sraw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32,        sext(32,        self.x[rs1]) >> (self.x[rs2]&31))
    def _fence    (self,                   **_): self.pc+=4
    def _fence_i  (self,                   **_): self.pc+=4
    def _csrrw    (self, rd, csr, rs1,     **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, self.x[rs1]               )
    def _csrrs    (self, rd, csr, rs1,     **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, self.csr[csr]| self.x[rs1])
    def _csrrc    (self, rd, csr, rs1,     **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, self.csr[csr]&~self.x[rs1])
    def _csrrwi   (self, rd, csr, zimm,    **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, zimm                      )
    def _csrrsi   (self, rd, csr, zimm,    **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, self.csr[csr]| zimm       )
    def _csrrci   (self, rd, csr, zimm,    **_): self.pc+=4; self.x[rd], self.csr[csr] = self.csr[csr], self.hook_csr(csr, self.csr[csr]&~zimm       )
    def _mret     (self,                   **_): self.pc = zext(self.xlen, self.csr[self.mepc]); mmode = (self.csr[self.mstatus]>>11)&3; self.csr[self.mstatus] = (self.current_mode << 11) | 0x80 | ((self.csr[self.mstatus]&0x80) >> 4); self.current_mode = mmode
    def _ecall    (self,                   **_): self.mtrap(0, 8 if self.current_mode == 0 else 11)
    def _ebreak   (self,                   **_): self.mtrap(self.op.addr, 3)
    def _mul      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (                self.x[rs1]  *                 self.x[rs2] )           )
    def _mulh     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (                self.x[rs1]  *                 self.x[rs2] )>>self.xlen)
    def _mulhu    (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (zext(self.xlen, self.x[rs1]) * zext(self.xlen, self.x[rs2]))>>self.xlen)
    def _mulhsu   (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (                self.x[rs1]  * zext(self.xlen, self.x[rs2]))>>self.xlen)
    def _div      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (self.idiv2zero( self.x[rs1]  ,                 self.x[rs2]))           ) if self.x[rs2] != 0 else -1
    def _divu     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (zext(self.xlen, self.x[rs1]) //zext(self.xlen, self.x[rs2]))           ) if self.x[rs2] != 0 else -1
    def _rem      (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (self.rem2zero ( self.x[rs1]  ,                 self.x[rs2]))           ) if self.x[rs2] != 0 else self.x[rs1]
    def _remu     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(self.xlen, (zext(self.xlen, self.x[rs1]) % zext(self.xlen, self.x[rs2]))           ) if self.x[rs2] != 0 else self.x[rs1]
    def _mulw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32,        (sext(32,        self.x[rs1]) * sext(32,        self.x[rs2]))           )
    def _divw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, (self.idiv2zero(sext(32,self.x[rs1]) , sext(32,        self.x[rs2])))          ) if sext(32,self.x[rs2]) != 0 else -1
    def _divuw    (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, (self.idiv2zero(zext(32,self.x[rs1]) , zext(32,        self.x[rs2])))          ) if sext(32,self.x[rs2]) != 0 else -1
    def _remw     (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, (self.rem2zero (sext(32,self.x[rs1]) , sext(32,        self.x[rs2])))          ) if sext(32,self.x[rs2]) != 0 else sext(32,self.x[rs1])
    def _remuw    (self, rd, rs1, rs2,     **_): self.pc+=4; self.x[rd] = sext(32, (self.rem2zero (zext(32,self.x[rs1]) , zext(32,        self.x[rs2])))          ) if sext(32,self.x[rs2]) != 0 else sext(32,self.x[rs1])
    def _amoswap_w(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1,                            sext(32,ors2)  ); self.x[rd]=tmp
    def _amoadd_w (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32,             tmp + sext(32,ors2)) ); self.x[rd]=tmp
    def _amoand_w (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32,             tmp & sext(32,ors2)) ); self.x[rd]=tmp
    def _amoor_w  (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32,             tmp | sext(32,ors2)) ); self.x[rd]=tmp
    def _amoxor_w (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32,             tmp ^ sext(32,ors2)) ); self.x[rd]=tmp
    def _amomax_w (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32, max(        tmp , sext(32,ors2)))); self.x[rd]=tmp
    def _amomin_w (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32, min(        tmp , sext(32,ors2)))); self.x[rd]=tmp
    def _amomaxu_w(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32, max(zext(32,tmp), zext(32,ors2)))); self.x[rd]=tmp
    def _amominu_w(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('i', self.x[rs1], self.x[rd]); self.store('i', ors1, sext(32, min(zext(32,tmp), zext(32,ors2)))); self.x[rd]=tmp
    def _amoswap_d(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1,                                    ors2,  ); self.x[rd]=tmp
    def _amoadd_d (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64,             tmp +         ors2 ) ); self.x[rd]=tmp
    def _amoand_d (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64,             tmp &         ors2 ) ); self.x[rd]=tmp
    def _amoor_d  (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64,             tmp |         ors2 ) ); self.x[rd]=tmp
    def _amoxor_d (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64,             tmp ^         ors2 ) ); self.x[rd]=tmp
    def _amomax_d (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64, max(        tmp , sext(64,ors2)))); self.x[rd]=tmp
    def _amomin_d (self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64, min(        tmp , sext(64,ors2)))); self.x[rd]=tmp
    def _amomaxu_d(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64, max(zext(64,tmp), zext(64,ors2)))); self.x[rd]=tmp
    def _amominu_d(self, rd, rs1, rs2,     **_): self.pc+=4; ors1, ors2 = self.x[rs1], self.x[rs2]; tmp = self.load('q', self.x[rs1], self.x[rd]); self.store('q', ors1, sext(64, min(zext(64,tmp), zext(64,ors2)))); self.x[rd]=tmp
    def _lr_w     (self, rd, rs1,          **_): self.pc+=4; self.lr_res_addr = self.x[rs1]; self.x[rd] = self.load('i', self.x[rs1], self.x[rd])
    def _sc_w     (self, rd, rs1, rs2,     **_):
        if self.lr_res_addr == self.x[rs1]: self.pc+=4; self.lr_res_addr = -1; self.store('i', self.x[rs1], sext(32, self.x[rs2])); self.x[rd] = 0
        else:                               self.pc+=4; self.lr_res_addr = -1;                                                      self.x[rd] = 1
    def _c_addi  (self, rd_rs1_n0, nzimm6, **_): self.pc+=2; self.x[rd_rs1_n0] += nzimm6  # c.nop required to pass test rv32i_m/privilege/src/misalign-jal-01.S
    def hook_exec(self): return True
    def unimplemented(self, **_): print(f'\n{zext(64,self.op.addr):08x}: unimplemented: {zext(32,self.op.data):08x} {self.op}')
    def step(self, trace=True):
        self.op = decode(self.load('I', self.pc, notify=False, check_misaligned=False), addr=self.pc)
        self.cycle += 1; self.csr[self.mcycle] = zext(self.xlen, self.cycle)
        self.trace_log = [] if trace else None
        if self.hook_exec():
            getattr(self, '_'+self.op.name, self.unimplemented)(**self.op.args)  # dynamic instruction dispatch
            if trace: print(f'{zext(64,self.op.addr):08x}: {str(self.op):40} #', ' '.join(self.trace_log))
            if trace and self.pc-self.op.addr not in (2, 4): print()
    def run(self, limit=0, bpts=set(), trace=True):
        while True:
            self.step(trace=trace)
            if self.op.addr in bpts|{self.pc} or (limit := limit-1)==0: break